import json
from typing import Any, Dict

from aiohttp import BasicAuth, ClientConnectionError, ClientSession

from .const import _LOGGER, HTTP_OK, HTTP_UNAUTHORIZED


class BaseAPI:
    """Class to perform CMI API requests."""

    def __init__(
        self, username: str, password: str, session: ClientSession = None
    ) -> None:
        """Initialize."""
        self.auth = BasicAuth(username, password)
        self.session = session

        self._internal_session = False

    async def _make_request(self, url: str) -> Dict[str, Any]:
        """Retrieve data from CMI API."""
        raw_response: str = await self._make_request_no_json(url)
        data = json.loads(raw_response)

        STATUS_CODE = "Status code"

        if data[STATUS_CODE] != 0:
            _LOGGER.debug("Failed response received: %s", raw_response)

        if data[STATUS_CODE] == 0:
            return data
        elif data[STATUS_CODE] == 1:
            raise ApiError("Node not available")
        elif data[STATUS_CODE] == 2:
            raise ApiError(
                "Failure during the CAN-request/parameter not available for this device"
            )
        elif data[STATUS_CODE] == 4:
            raise RateLimitError("Only one request per minute is permitted")
        elif data[STATUS_CODE] == 5:
            raise ApiError("Device not supported")
        elif data[STATUS_CODE] == 7:
            raise ApiError("CAN Bus is busy")
        else:
            raise ApiError("Unknown error")

    @staticmethod
    def _fix_invalid_json(invalid_json: str) -> str:
        """Fix invalid json provided by modbus parameter."""
        return invalid_json.replace("{ ,", "{")

    async def _make_request_no_json(self, url: str) -> str:
        """Retrieve data from CMI API that is not valid json."""
        if self.session is None:
            self._internal_session = True
            self.session = ClientSession()

        try:
            async with self.session.get(url, auth=self.auth) as res:
                await self._close_session()
                if res.status == HTTP_UNAUTHORIZED:
                    raise InvalidCredentialsError("Invalid API key")
                elif res.status != HTTP_OK:
                    raise ApiError(f"Invalid response from CMI: {res.status}")

                text = self._fix_invalid_json(await res.text())
                return text
        except ClientConnectionError:
            await self._close_session()
            raise ApiError(f"Could not connect to C.M.I")

    async def _close_session(self) -> None:
        """Close the internal session."""
        if self._internal_session:
            await self.session.close()
            self.session = None


class ApiError(Exception):
    """Raised when API request ended in error."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status


class InvalidCredentialsError(Exception):
    """Triggered when the credentials are invalid."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status


class RateLimitError(Exception):
    """Triggered when the rate limit is reached."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status



from typing import Any, Dict, List

from aiohttp import ClientSession

from .baseApi import BaseAPI
from .channel import Channel
from .const import _LOGGER, DEVICES, SUPPORTED_PARAMS_FOR_DEVICE, ChannelType


class Device(BaseAPI):
    """Class to interact with a device."""

    def __init__(
        self,
        node_id: str,
        host: str,
        username: str,
        password: str,
        session: ClientSession = None,
    ) -> None:
        """Initialize."""
        super().__init__(username, password, session)
        self.id: str = node_id
        self.host: str = host

        self.api_version: int = 0
        self.device_id: str = "00"

        self._channels: Dict[ChannelType, Dict[int, Channel]] = {}

    def _extract_device_info(self, json: Dict[str, Any]) -> None:
        """Extract device info from request response."""
        self.api_version: int = json["Header"]["Version"]
        self.device_id: str = json["Header"]["Device"]

        if self.device_id not in DEVICES.keys():
            raise InvalidDeviceError(f"Invalid device id: {self.device_id}")

    @staticmethod
    def _extract_channels(
        mode: ChannelType, raw_channels: List[Dict[str, Any]]
    ) -> Dict[int, Channel]:
        """Extract channel info from data array from request."""
        list_of_channels: Dict[int, Channel] = {}
        for channel_raw in raw_channels:
            ch: Channel = Channel(mode, channel_raw)
            list_of_channels[ch.index] = ch

        return list_of_channels

    def _get_json_params(self) -> str:
        """Compose json params based on the device type."""
        default_params = "I,O"
        return SUPPORTED_PARAMS_FOR_DEVICE.get(self.device_id, default_params)

    async def _make_request_to_device(self) -> Dict[str, Any]:
        """"""
        url: str = f"{self.host}/INCLUDE/api.cgi?jsonparam={self._get_json_params()}&jsonnode={self.id}"
        _LOGGER.debug(
            "Make request to device %s from type %s with parameters: %s",
            self.id,
            self.device_id,
            self._get_json_params(),
        )
        return await self._make_request(url)

    async def fetch_type(self) -> None:
        """Fetch the device type without parsing the data from the device."""
        self._extract_device_info(await self._make_request_to_device())

    async def update(self) -> None:
        """Update data."""
        _LOGGER.debug("Update device: %s", self.id)
        res: Dict[str, Any] = await self._make_request_to_device()

        if self.device_id == "00":
            self._extract_device_info(res)
            _LOGGER.debug("Device had no id. Set new id to %s", self.device_id)

        for channel_type_text in res["Data"]:
            channel_type = ChannelType(channel_type_text)
            self._channels[channel_type] = self._extract_channels(
                channel_type, res["Data"][channel_type_text]
            )

    def get_channels(self, channel_type: ChannelType) -> Dict[int, Channel]:
        """Get all the fetched channels from a type."""
        return self._channels[channel_type]

    def has_channel_type(self, channel_type: ChannelType) -> bool:
        """Check if a channel type was fetched."""
        return self._channels.get(channel_type, None) is not None

    def set_device_type(self, device_name: str) -> None:
        """Set the type of the device manually."""
        type_id = [i for i in DEVICES if DEVICES[i] == device_name]

        if len(type_id) != 1:
            raise InvalidDeviceError(f"Invalid device name: {device_name}")

        self.device_id = type_id[0]

    def get_device_type(self) -> str:
        """Get the type of the device."""
        return DEVICES.get(self.device_id, "Unknown")

    def __repr__(self) -> str:
        text = f"Node {self.id}: Type: {self.get_device_type()}"

        for channel_type in self._channels:
            text += f", {channel_type.value}: {len(self._channels[channel_type])}"

        return text


class InvalidDeviceError(Exception):
    """Triggered when an invalid device type is set."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status



from typing import Any, Dict

from .const import (
    UNITS_DE,
    UNITS_EN,
    ChannelMode,
    ChannelType,
    Languages,
    ReadOnlyClass,
)


class Channel(metaclass=ReadOnlyClass):
    """Class to display an input or output."""

    def __init__(self, mode: ChannelType, json: Dict[str, Any]) -> None:
        """Initialize and parse json to get properties."""
        self.mode: ChannelType = mode
        self.type: ChannelMode = json["AD"]
        self.index: int = json["Number"]
        self.value: float = json["Value"]["Value"]
        self.unit: str = json["Value"]["Unit"]

    def get_unit(self, language: Languages = Languages.EN) -> str:
        """Get the unit of the channel."""
        if language == Languages.EN:
            return UNITS_EN.get(self.unit, "Unknown")
        else:
            return UNITS_DE.get(self.unit, "Unknown")

    def __eq__(self, other) -> bool:
        return (
            self.mode == other.mode
            and self.type == other.type
            and self.index == other.index
            and self.value == other.value
            and self.unit == other.unit
        )

    def __repr__(self) -> str:
        return f"Channel {self.index}: Type: {self.type}, Mode: {self.mode}, Value: {self.value} {self.get_unit()}"



__version__ = "1.3.1"
__author__ = "DeerMaximum"

from .baseApi import ApiError, InvalidCredentialsError, RateLimitError
from .channel import Channel
from .cmi import CMI
from .const import ChannelMode, ChannelType, Languages
from .device import Device, InvalidDeviceError



from typing import List

from aiohttp import ClientSession

from .baseApi import BaseAPI
from .const import ReadOnlyClass
from .device import _LOGGER, Device


class CMI(BaseAPI, metaclass=ReadOnlyClass):
    """Main class to interact with CMI."""

    def __init__(
        self, host: str, username: str, password: str, session: ClientSession = None
    ) -> None:
        """Initialize."""
        super().__init__(username, password, session)
        self.host = host
        self.username = username
        self.password = password

    async def get_devices(self) -> List[Device]:
        """List connected devices."""
        _LOGGER.debug("Receive list of nodes from C.M.I")
        url: str = f"{self.host}/INCLUDE/can_nodes.cgi?_=1"
        data: str = await self._make_request_no_json(url)

        _LOGGER.debug("Received list of nodes from C.M.I: %s", data)

        node_ids: List[str] = data.split(";")

        devices: List[Device] = []

        for node_id in node_ids:
            if len(node_id) == 0:
                continue
            devices.append(
                Device(node_id, self.host, self.username, self.password, self.session)
            )

        return devices
